"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.assert_return = assert_return;
exports.assert_malformed = assert_malformed;
exports.assert_invalid = assert_invalid;
exports.assert_trap = assert_trap;

var _long = require("@xtuc/long");

var _long2 = _interopRequireDefault(_long);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function eq(actual, expected) {

  // check value
  if (!(actual.type === expected.type)) {
    throw new Error('actual.type === expected.type' + " error: " + ("type mismatch; expected " + expected.type + ", given " + actual.type || "unknown"));
  }
  // check type
  switch (expected.type) {
    case "i32":
      {
        var i32Value = _long2.default.fromString(expected.value).toInt();

        if (!(actual.value.toString() === i32Value.toString())) {
          throw new Error('actual.value.toString() === i32Value.toString()' + " error: " + ("Expected value " + i32Value + ", got " + actual.value.toString() || "unknown"));
        }
        break;
      }

    case "f32":
      {
        var actuali32 = actual.value.reinterpret();
        var expectedi32 = _long2.default.fromString(expected.value).toInt();

        if (!(actuali32.toNumber() === expectedi32)) {
          throw new Error('actuali32.toNumber() === expectedi32' + " error: " + ("Expected value " + expectedi32 + ", got " + actuali32.toString() || "unknown"));
        }

        break;
      }

    case "f64":
      {
        var _actuali = actual.value.reinterpret();
        var _expectedi = _long2.default.fromString(expected.value).toNumber();

        if (!(_actuali.toNumber() === _expectedi)) {
          throw new Error('_actuali.toNumber() === _expectedi' + " error: " + ("Expected value " + _expectedi + ", got " + _actuali.toString() || "unknown"));
        }

        break;
      }

    case "i64":
      {
        console.warn("eq with i64 is unsupported");
        // const actuali64 = actual.value.toString();
        // const expectedi64 = Long.fromString(expected.value)
        //   .toSigned()
        //   .toString();

        // FIXME(sven): fix this
        // assert(
        //   actuali64 === expectedi64,
        //   `Expected value ${expectedi64}, got ${actuali64}`
        // );
        break;
      }

    default:
      throw new Error("Unsupport eq with type: " + expected.type);
  }
}

// assert action has expected results
// ( assert_return <action> <expr>* )
//
function assert_return(element, action, expected) {
  var type = action.type,
      args = action.args;
  if (!(type === "invoke" || type === "get")) {
    throw new Error('type === "invoke" || type === "get"' + " error: " + ("unsupported type \"" + type + "\"" || "unknown"));
  }

  if (type === "get") {
    if (expected.length > 0) {
      eq(element, expected[0]);
    }
  }

  if (type === "invoke") {
    var compatibleArgs = args.map(function (x) {
      if (x.type === "i64") {
        return new _long2.default.fromString(x.value);
      }

      return x.value;
    });

    var res = element.apply(undefined, _toConsumableArray(compatibleArgs));

    if (expected.length > 0) {
      eq(res, expected[0]);
    }
  }
}

// ;; assert module cannot be decoded with given failure string
// ( assert_malformed <module> <failure> )
//
function assert_malformed(getInstance, expected) {
  try {
    getInstance();

    if (!false) {
      throw new Error('false' + " error: " + ("did not throw any error" || "unknown"));
    }
  } catch (e) {
    if (!e.message.match(new RegExp(expected, "gm"))) {
      throw new Error('e.message.match(new RegExp(expected, "gm"))' + " error: " + ("Expected error \"" + expected + "\", got \"" + e.message + "\"" || "unknown"));
    }
  }
}

// assert module is invalid with given failure string
// ( assert_invalid <module> <failure> )
//
function assert_invalid(getInstance, expected) {
  if (expected === "type mismatch") {
    expected = "Expected type|Stack contains additional type";
  }

  try {
    getInstance();

    if (!false) {
      throw new Error('false' + " error: " + ("did not throw any error" || "unknown"));
    }
  } catch (e) {
    if (!e.message.match(new RegExp(expected, "gm"))) {
      throw new Error('e.message.match(new RegExp(expected, "gm"))' + " error: " + ("Expected error \"" + expected + "\", got \"" + e.message + "\"" || "unknown"));
    }
  }
}

// assert module traps on instantiation
// ( assert_trap <module> <failure> )
//
function assert_trap(element, action, expected) {
  var type = action.type,
      args = action.args;
  if (!(type === "invoke")) {
    throw new Error('type === "invoke"' + " error: " + ("unsupported type \"" + type + "\"" || "unknown"));
  }

  if (type === "invoke") {
    var compatibleArgs = args.map(function (x) {
      if (x.type === "i64") {
        return new _long2.default.fromString(x.value);
      }

      return x.value;
    });

    try {
      element.apply(undefined, _toConsumableArray(compatibleArgs));
    } catch (e) {
      if (!e.message.match(new RegExp(expected, "gm"))) {
        throw new Error('e.message.match(new RegExp(expected, "gm"))' + " error: " + ("Expected error \"" + expected + "\", got \"" + e.message + "\"" || "unknown"));
      }
    }
  }
}