"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.stack = exports.isConst = exports.typeEq = exports.getType = undefined;
exports.default = validateAST;
exports.getValidationErrors = getValidationErrors;

var _typeInference = require("./type-inference");

Object.defineProperty(exports, "getType", {
  enumerable: true,
  get: function get() {
    return _typeInference.getType;
  }
});
Object.defineProperty(exports, "typeEq", {
  enumerable: true,
  get: function get() {
    return _typeInference.typeEq;
  }
});

var _importOrder = require("./import-order");

var _importOrder2 = _interopRequireDefault(_importOrder);

var _isConst = require("./is-const");

var _isConst2 = _interopRequireDefault(_isConst);

var _typeChecker = require("./type-checker");

var _typeChecker2 = _interopRequireDefault(_typeChecker);

var _imports = require("./imports");

var _imports2 = _interopRequireDefault(_imports);

var _duplicatedExports = require("./duplicated-exports");

var _duplicatedExports2 = _interopRequireDefault(_duplicatedExports);

var _helperModuleContext = require("@webassemblyjs/helper-module-context");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function validateAST(ast) {
  var errors = getValidationErrors(ast);

  if (errors.length !== 0) {
    var errorMessage = "Validation errors:\n" + errors.join("\n");

    throw new Error(errorMessage);
  }
}

function getValidationErrors(ast) {
  var errors = [];

  var modules = [];

  // $FlowIgnore
  if (ast.type === "Module") {
    modules = [ast];
  }

  // $FlowIgnore
  if (ast.type === "Program") {
    modules = ast.body.filter(function (_ref) {
      var type = _ref.type;
      return type === "Module";
    });
  }

  modules.forEach(function (m) {
    var moduleContext = (0, _helperModuleContext.moduleContextFromModuleAST)(m);

    // $FlowIgnore
    errors.push.apply(errors, _toConsumableArray((0, _imports2.default)(ast, moduleContext)));
    errors.push.apply(errors, _toConsumableArray((0, _isConst2.default)(ast, moduleContext)));
    errors.push.apply(errors, _toConsumableArray((0, _importOrder2.default)(ast)));
    errors.push.apply(errors, _toConsumableArray((0, _typeChecker2.default)(ast, moduleContext)));
    errors.push.apply(errors, _toConsumableArray((0, _duplicatedExports2.default)(ast)));
  });

  return errors;
}

exports.isConst = _isConst2.default;
var stack = exports.stack = _typeChecker2.default;